/*
 * env-sensor.c
 *
 * Copyright (c) 2012, ROSAND Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the Contiki operating system.
 *
 */

#include <stdint.h>
#include "sys/clock.h"
#include "defs.h"
#include "dev/env-sensor.h"
#include "dev/adc.h"
#include "dev/i2c.h"
#include "dev/leds.h"

#define I2C_ADDRESS             0x27    /* Honeywell HIH-6130/6131 I2C default address */
#define I2C_CYCLE               40      /* Cycle measurement Delay in milliseconds */
#define I2C_LENGTH              0x04    /* Length of sensor data */

/* TODO: Measure VD and change I2C_LIGHT_VD by parameter stored in Flash/EEPROM */
#define I2C_LIGHT_VD            0x00    /* Voltage read in the absence of light */
#define I2C_REPONSIVITY         15.6    /* Responsivity Re (= Irradiance / Voltage) */

/*-------------------------------------------------------------------------------
 * fetch()
 * @brief       Fetch data from the HIH-6130/6131 sensor
 */
static int
fetch(uint8_t sensor)
{
  uint8_t buf[I2C_LENGTH];
  uint8_t status;
  uint16_t value = 0;

  i2c_init(I2C_MASTER);                         /* Put I2C is Master mode */
  i2c_receive(I2C_ADDRESS, buf, I2C_LENGTH);    /* Fetch sensor data */
  i2c_init(I2C_SLAVE);                          /* Revert back I2C to default Slave mode */
  status = buf[0] >> 6;                         /* Status bits are the 2 MSB of the first data byte */

  if (status & 0x02) {                          /* Device in Command Mode or Diagnostic Condition */
    return -1;                                  /* Ignore the data */
  }

  buf[0] &= 0x3F;                               /* Remove status bits */
  switch(sensor) {
  case ENV_HUMID:
    value = BUILD_UINT16(buf[1], buf[0]);       /* The two fist bytes are humidity */
    break;
  case ENV_TEMP:
    value = BUILD_UINT16(buf[3], buf[2]);       /* The last two bytes are temperature */
    value >>= 2;                                /* Ignore the 2 LSB, value is 14-bit long */
    break;
  default:
    value = -1;
  }
  return value;
}

/*-------------------------------------------------------------------------------
 * value()
 * @brief       Return environment sensor value:
 *              Temperature in 100th of degrees
 *              Humidity in % RH
 *              Irradiance in microWatt/cm²
 */
static int
value(int type)
{
  int16_t res;
  float val;

  switch (type) {
  case ENV_HUMID:
    res = fetch(type);
    if (res == -1) {
      return -1;
    }
    /*
     * Humidity Conversion:
     * The output of the device is a 14-bit number representing
     * between 0 %RH and 100 %RH.
     *          H (%RH) = (val / (power(2, 14) - 1)) * 100%
     */
    val = res;
    val = val * 100/((1 << 14) -1);
    return val;
  case ENV_TEMP:
    res = fetch(type);
    if (res == -1) {
      return -1;
    }
    /*
     * Temperature Conversion:
     * The output of the device is a 14-bit number representing
     * between -40°C and 125°C.
     *          T (°C) = (val / ((power(2, 14) - 1) * 165) - 40
     */
    val = res;
    val = val * 165/((1 << 14) -1) - 40;
    return val * 100;           /* return 100th degrees */
  case ENV_LUX:
    res = adc(ENV_LUX_CHANNEL);
    if (res == -1) {
      return -1;
    }
    /*
     * Light Conversion (TAOS TSL250RD):
     * The output of the device is a 10-bit number representing
     * between VD (voltage in the absence of light) and VOM (voltage
     * output for Vref).
     *          V = (VOM / (power(2, 10) - 1)) + VD
     * According to figure 6 of the datasheet:
     *          VOM = 2 + 2 * (Vref - 3) / 4.4
     *          VOM = (Vref + 1.4) / 2.2
     * Vref = 3.3 V, hence
     *          VOM = 4.7 / 2.2
     *          V = 4.7 / (2.2 * (power(2, 10) - 1)) + VD
     *          res = V - VD
     * According to figure 2 of the datasheet:
     *          Ee = (V - VD) * Re (uW/cm2)
     *          Ee = 4.7 * Re / (2.2 * power(2, 10) - 1))
     * VD being I2C_LIGHT_VD, Re being I2C_REPONSIVITY, we finally have:
     */
    val  = res - I2C_LIGHT_VD;
    val *= 4.7 * I2C_REPONSIVITY / (2.2 * ((1 << 10) -1));
    return val;
  default:
    return -1;
  }
}

/*-------------------------------------------------------------------------------
 * configure()
 */
static int
configure(int type, int c)
{
  switch (type) {
  case SENSORS_ACTIVE:
    if (c) {
      return 1;
    }
    else {
      return 0;
    }
  default:
    return 0;
  }
}

/*-------------------------------------------------------------------------------
 * status()
 */
static int
status(int type)
{
  switch (type) {
  case SENSORS_ACTIVE:
  case SENSORS_READY:
    return 1;
  default:
    return 0;
  }
}

SENSORS_SENSOR(env_sensor, "Environment Sensor", value, configure, status);
