/*
 * i2c.c
 *
 * Copyright (c) 2012, ROSAND Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the Contiki operating system.
 *
 */

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "contiki.h"
#include "dev/leds.h"
#include "dev/i2c.h"
#include "defs.h"

#ifndef SCLPORT
#define SCLPORT    D
#endif /* SCLPORT */
#ifndef SCLPIN
#define SCLPIN     0
#endif /* CLKPIN */
#ifndef SDAPORT
#define SDAPORT    D
#endif /* SDAPORT */
#ifndef SDAPIN
#define SDAPIN     1
#endif /* SDAPIN*/

#define I2C_SCL_DDR     DDR(SCLPORT)
#define I2C_SCL_PORT    PORT(SCLPORT)
#define I2C_SCL_PIN     SCLPIN
#define I2C_SDA_DDR     DDR(SDAPORT)
#define I2C_SDA_PORT    PORT(SDAPORT)
#define I2C_SDA_PIN     SDAPIN


/****************************************************************************
  I2C State codes
****************************************************************************/
// General TWI Master status codes
#define I2C_START                  0x08  // START has been transmitted
#define I2C_REP_START              0x10  // Repeated START has been transmitted
#define I2C_ARB_LOST               0x38  // Arbitration lost

/* I2C Master Transmitter status codes */
#define I2C_MTX_ADR_ACK            0x18  // SLA+W has been tramsmitted and ACK received
#define I2C_MTX_ADR_NACK           0x20  // SLA+W has been tramsmitted and NACK received
#define I2C_MTX_DATA_ACK           0x28  // Data byte has been tramsmitted and ACK received
#define I2C_MTX_DATA_NACK          0x30  // Data byte has been tramsmitted and NACK received

/* I2C Master Receiver status codes */
#define I2C_MRX_ADR_ACK            0x40  // SLA+R has been tramsmitted and ACK received
#define I2C_MRX_ADR_NACK           0x48  // SLA+R has been tramsmitted and NACK received
#define I2C_MRX_DATA_ACK           0x50  // Data byte has been received and ACK tramsmitted
#define I2C_MRX_DATA_NACK          0x58  // Data byte has been received and NACK tramsmitted

/* I2C Slave Transmitter status codes */
#define I2C_STX_ADR_ACK            0xA8  // Own SLA+R has been received; ACK has been returned
#define I2C_STX_ADR_ACK_M_ARB_LOST 0xB0  // Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
#define I2C_STX_DATA_ACK           0xB8  // Data byte in TWDR has been transmitted; ACK has been received
#define I2C_STX_DATA_NACK          0xC0  // Data byte in TWDR has been transmitted; NOT ACK has been received
#define I2C_STX_DATA_ACK_LAST_BYTE 0xC8  // Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received

/* I2C Slave Receiver status codes */
#define I2C_SRX_ADR_ACK            0x60  // Own SLA+W has been received ACK has been returned
#define I2C_SRX_ADR_ACK_M_ARB_LOST 0x68  // Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
#define I2C_SRX_GEN_ACK            0x70  // General call address has been received; ACK has been returned
#define I2C_SRX_GEN_ACK_M_ARB_LOST 0x78  // Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
#define I2C_SRX_ADR_DATA_ACK       0x80  // Previously addressed with own SLA+W; data has been received; ACK has been returned
#define I2C_SRX_ADR_DATA_NACK      0x88  // Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
#define I2C_SRX_GEN_DATA_ACK       0x90  // Previously addressed with general call; data has been received; ACK has been returned
#define I2C_SRX_GEN_DATA_NACK      0x98  // Previously addressed with general call; data has been received; NOT ACK has been returned
#define I2C_SRX_STOP_RESTART       0xA0  // A STOP condition or repeated START condition has been received while still addressed as Slave

/* I2C Miscellaneous status codes */
#define I2C_NO_STATE               0xF8  // No relevant state information available; TWINT = 0
#define I2C_BUS_ERROR              0x00  // Bus error due to an illegal START or STOP condition

static uint8_t i2c_buf[I2C_BUFFER_SIZE];
static volatile uint8_t i2c_len;
static uint8_t i2c_mode;                        /* i2c_mode = MASTER or SLAVE */

static volatile union i2c_sreg sreg = {0};      /* I2C status register */

PROCESS_NAME(config_process);

/*--------------------------------------------------------------------*/
static void
i2c_start(void)
{
  TWCR = (1 << TWEN) |                          /* Enable TWI interface */
         (1 << TWIE) | (1 << TWINT);            /* Enable interrupt and clear the flag */
  if(i2c_mode == I2C_MASTER) {
    TWCR |= (1 << TWSTA);                       /* Initiate a START condition */
  } else {
    TWCR |= (1 << TWEA);                        /* Prepare the ACK next time the Slave is addressed */
  }
}
/*--------------------------------------------------------------------*/
static uint8_t
i2c_busy(void)
{
  return (TWCR & (1 << TWIE));                  /* If Interrupt is enabled then the transceiver is busy */
}
/*--------------------------------------------------------------------*/
void
i2c_init(uint8_t mode)
{
  i2c_mode = mode;
  while(i2c_busy());                            /* Wait for current transfer to be completed */
  TWDR = 0xFF;                                  /* Default content = SDA released */
  TWCR = 0;                                     /* Disable TWI */
  switch(mode) {
  case I2C_MASTER:
    TWBR = 0x20;                                /* I2C frequency = 100 KHz with Clock frequency = 8 MHz */
    TWSR = 0x00;
    break;
  case I2C_SLAVE:
    TWAR = (I2C_P8_ADDRESS << 1);               /* Set TWI slave address */
    break;
  }
}
/*--------------------------------------------------------------------*/
static void
i2c_send_address(uint8_t address)
{
  i2c_buf[0]  = (address << 1);               /* Shift 7-bit address one bit left */
  i2c_buf[0] |= I2C_CMD_READ;                 /* Set direction bit read=1, write=0 */
}
/*--------------------------------------------------------------------*/
uint8_t
i2c_send(uint8_t address, uint8_t *data, uint8_t len)
{
  uint8_t i;
  uint8_t *tmp = i2c_buf;

  while (i2c_busy());                           /* Wait until I2C is ready for next transmission */

  if (len > I2C_BUFFER_SIZE) {
    len = I2C_BUFFER_SIZE;                      /* Limit to buffer size */
  }

  if(i2c_mode == I2C_MASTER) {
    *tmp++  = (address << 1);                   /* Shift 7-bit address one bit left */
  }
  for (i=0; i < len; i++) {
    *tmp++ = *data++;                           /* Store data block */
  }
  sreg.all = 0;                                 /* Clear status register */
  i2c_len = (tmp - i2c_buf);                    /* Length of data to transmit */
  i2c_start();
  return len;
}
/*--------------------------------------------------------------------*/
uint8_t
i2c_receive(uint8_t address, uint8_t *data, uint8_t len)
{
  uint8_t i = 0;

  if (len > I2C_BUFFER_SIZE) {
    len = I2C_BUFFER_SIZE;                      /* Limit to buffer size */
  }

  while(i2c_busy());                            /* Wait for end of reception */
  if(i2c_mode == I2C_MASTER) {                  /* MASTER mode */
    i2c_len = len;                              /* Number of data to receive */
    i2c_send_address(address);                  /* Send read request to slave */
  }
  i2c_start();

  while(i2c_busy());                            /* Wait for end of reception */
  if (sreg.u.ok) {
    for (i = 0; i < i2c_len; i++) {
      *(data + i) = i2c_buf[i];
    }
  }

  sreg.all = 0;                                 /* Clear status register */
  return i;
}
/*--------------------------------------------------------------------*/
ISR(TWI_vect)
{
  static uint8_t index;

  switch (TWSR) {
  case I2C_START:                               /* START has been transmitted */
  case I2C_REP_START:                           /* Repeated START has been transmitted */
    index = 0;                                  /* Set buffer pointer to the TWI Address location */
  case I2C_MTX_ADR_ACK:                         /* SLA+W has been tramsmitted and ACK received */
  case I2C_MTX_DATA_ACK:                        /* Data byte has been tramsmitted and ACK received */
    if (index < i2c_len) {                      /* Detect the last byte to send */
      TWDR = i2c_buf[index++];                  /* Send data and update index */
      TWCR = (1 << TWEN) |                      /* TWI Interface enabled */
          (1 << TWIE) | (1 << TWINT);           /* Enable TWI Interupt and clear the flag to send byte */
    } else {                                    /* Send STOP after last byte */
      sreg.u.ok = 1;                            /* Set status bits to completed successfully */
      TWCR = 0;                                 /* Release TWI pins */
      TWCR = (1 << TWEN) |                      /* TWI Interface enabled */
          (1 << TWINT) |                        /* Disable TWI Interrupt and clear the flag */
          (1 << TWSTO);                         /* Initiate a STOP condition. */
    }
    break;
  case I2C_MRX_ADR_ACK:                         /* SLA+R has been tramsmitted and ACK received */
    index = 0;                                  /* Clear index to start reception */
    TWCR = (1 << TWEN) |                        /* TWI Interface enabled */
        (1 << TWIE) |(1 << TWINT) |             /* Enable TWI Interupt and clear the flag to read next byte */
        (1 << TWEA);                            /* Send ACK after reception */
    break;
  case I2C_MRX_DATA_ACK:                        /* Data byte has been received and ACK tramsmitted */
    i2c_buf[index++] = TWDR;                    /* Store received data and update index */
    if (index < (i2c_len-1)) {                  /* Detect the last byte to NACK it. */
      TWCR = (1 << TWEN) |                      /* TWI Interface enabled */
          (1 << TWIE) |(1 << TWINT) |           /* Enable TWI Interupt and clear the flag to read next byte */
          (1 << TWEA);                          /* Send ACK after reception */
    } else  {                                   /* Send NACK after next reception */
      TWCR = (1 << TWEN) |                      /* TWI Interface enabled, send NACK after reception */
          (1 << TWIE) | (1 << TWINT);           /* Enable TWI Interupt and clear the flag to read next byte */
    }
    break;
  case I2C_MRX_DATA_NACK:                       /* Data byte has been received and NACK tramsmitted */
    i2c_buf[index] = TWDR;
    sreg.u.ok = 1;                              /* Set status bits to completed successfully */
    TWCR = (1 << TWEN )|                        /* TWI Interface enabled */
        (1 << TWINT) |                          /* Disable TWI Interrupt and clear the flag */
        (1 << TWSTO);                           /* Initiate a STOP condition. */
    break;
  case I2C_ARB_LOST:                            /* Arbitration lost */
    i2c_start();                                /* Enable TWI with interrupt to wait for a new event. */
    break;
  case I2C_STX_ADR_ACK:                         /* Own SLA+R has been received; ACK has been returned */
    index = 0;                                  /* Set buffer pointer to first data location */
  case I2C_STX_DATA_ACK:                        /* Data byte in TWDR has been transmitted; ACK has been received */
    TWDR = i2c_buf[index++];
    i2c_start();                                /* Enable TWI with interrupt to wait for a new event. */
    break;
  case I2C_STX_DATA_NACK:                       /* Data byte in TWDR has been transmitted; NACK has been received. */
    sreg.u.ok = 1;                              /* Set flag transmission successful */
    TWCR = (1 << TWEN) | (1 << TWINT);          /* TWI Interface enabled, pins released */
    break;
  case I2C_SRX_GEN_ACK:                         /* General call address has been received; ACK has been returned */
    sreg.u.gen = 1;
  case I2C_SRX_ADR_ACK:                         /* Own SLA+W has been received ACK has been returned */
    sreg.u.in = 1;
    i2c_len = 0;                             /* Set buffer pointer to first data location */
    i2c_start();                                /* Enable TWI with interrupt to wait for a new event. */
     break;
  case I2C_SRX_ADR_DATA_ACK:       /* Previously addressed with own SLA+W; data has been received; ACK has been returned */
  case I2C_SRX_GEN_DATA_ACK:       /* Previously addressed with general call; data has been received; ACK has been returned */
    i2c_buf[i2c_len++] = TWDR;
    sreg.u.ok = 1;                              /* Set flag transmission successful */
    i2c_start();                                /* Enable TWI with interrupt to wait for a new event. */
    break;
  case I2C_SRX_STOP_RESTART:       /* A STOP condition or repeated START condition has been received while still addressed as Slave */
                                                /* Put TWI Transceiver in passive mode. */
    TWCR = (1 << TWEN);                         /* TWI Interface enabled */
    break;
  default:
    /* Rest TWI-interface */
    TWCR = (1 << TWEN);                         /* TWI Interface enabled */
  }
}
