/*
 * spi-arch.c
 *
 * Copyright (c) 2012, ROSAND Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the Contiki operating system.
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "contiki-net.h"
#include "dev/spi.h"
#include "dev/spi-arch.h"

PROCESS_NAME(rmote_radio_process);

static uint8_t data[BUFFER_SIZE];
static volatile uint8_t rx_head;                /* rx_head is modified both inside and outside the ISR */
static uint8_t rx_tail;
static volatile uint8_t receiving;              /* flag which notifies that data is being received */

/*-------------------------------------------------------------------------*/
static void
spi_setup(uint8_t type)
{
  switch(type) {
  case SPI_MASTER:
    while ((HAL_PIN_SS & (1 << HAL_SS_PIN)) == 0);              /* Wait until SPI ready */

    /* Set MOSI and CLK as output, all others input. */
    /* To avoid a SPI glitch, the SCK port register shall be set before the DDR register */
    HAL_PORT_SPI |= (1 << HAL_DD_SCK);                          /* Set CLK high */
    HAL_DDR_SPI  = (1 << HAL_DD_MOSI) | (1 << HAL_DD_SCK);      /* MOSI, CLK output */
    /* Set SS as output */
    HAL_DDR_SS  |= (1 << HAL_DD_SS);                            /* Set SS output */
    SPCR  = (1 << SPE) | (1 << MSTR);                           /* Enable SPI module in master mode, no interrupt */
    SPCR |= (1 << SPR1);                                        /* Run SPI at clock / 64 */
    SPSR  = (1 << SPI2X);                                       /* Double SPI speed => clock / 32 */
    HAL_PORT_SS &= ~(1 << HAL_DD_SS);                           /* Set SS low */
    break;
  case SPI_SLAVE:
    HAL_PORT_SS |= (1 << HAL_DD_SS);                            /* Set SS high */
    /* Set MISO as output, all others input. */
    HAL_DDR_SPI = (1 << HAL_DD_MISO);                           /* MISO output */
    SPCR &= ~(1 << MSTR);                                       /* Enable SPI module in slave mode */
    SPCR |= (1 << SPIE);                                        /* Enable SPI interrupt */
    break;
  }
}
/*-------------------------------------------------------------------------*/
void
spi_flush(void)
{
  rx_head = rx_tail = 0;                                        /* Clear queue buffer indexes */
  receiving = 0;                                                /* Idle state */
}
/*-------------------------------------------------------------------------*/
void
spi_init(void)
{
  spi_flush();                                                  /* Flush SPI buffer */

  /* Set MOSI and CLK as output, all others input. */
  /* To avoid a SPI glitch, the SCK port register shall be set before the DDR register */
  HAL_PORT_SPI |= (1 << HAL_DD_SCK);                            /* Set CLK high */
  HAL_DDR_SPI  |= (1 << HAL_DD_MOSI) | (1 << HAL_DD_SCK);        /* MOSI, CLK output */
  SPCR  = (1 << SPE) | (1 << MSTR);                             /* Enable SPI module in master mode, no interrupt */
  spi_setup(SPI_SLAVE);

  /* Enable Pin Change Interrupt on SS pin */
  /* SS pin is PCINT0 */
  PCICR  |= (1 << PCIE0);                                       /* Enable interrupt on PCINT7:0 */
  PCMSK0 |= (1 << PCINT0);                                      /* Set PCINT0 mask */
}
/*-------------------------------------------------------------------------*/
uint8_t
spi_read(uint8_t *buf, uint8_t len)
{
  uint8_t i = 0;
  uint8_t head = rx_head;

  while(((head - rx_tail) & BUFFER_MASK) > 0) {
    *(buf + i) = data[rx_tail];
    rx_tail = (rx_tail + 1) & BUFFER_MASK;
    i++;
    if(i == len) {
      break;
    }
  }
  return i;
}
/*-------------------------------------------------------------------------*/
uint8_t
spi_write(uint8_t *buf, uint8_t len)
{
  uint8_t i = 0;

  spi_setup(SPI_MASTER);
  while (len--) {
    SPI_WRITE(buf[i]);
    SPDR;
    i++;
  }
  spi_setup(SPI_SLAVE);
  return i;
}
/*-------------------------------------------------------------------------*/
ISR(SPI_STC_vect)
{
  uint8_t c;

  while(SPSR & (1 << WCOL));                                    /* Waiting for byte transfer to complete */
  c = SPDR;                                                     /* Read byte transfered */
  if(((rx_head - rx_tail) & BUFFER_MASK) == BUFFER_MASK) {      /* If no available space in buffer */
    receiving = 0;                                              /* Stop receiving packet */
    process_poll(&rmote_radio_process);                         /* Run SPI process */
  } else {
    receiving = 1;                                              /* We are receiving packet */
    data[rx_head] = c;                                          /* Store received byte to buffer */
    rx_head = (rx_head + 1) & BUFFER_MASK;                      /* Update new index */
  }
}
/*-------------------------------------------------------------------------*/
ISR(PCINT0_vect)
{
  if(receiving && (HAL_PIN_SS & _BV(HAL_SS_PIN))) {             /* SSN high, SPI transfer completed */
    receiving = 0;                                              /* Receiving packet finished */
    process_poll(&rmote_radio_process);                         /* Run SPI process */
  }
}
/*-------------------------------------------------------------------------*/
